package com.interdataworking.gateway;

import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import com.interdataworking.*;
import com.interdataworking.util.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;

/**
 * By default, all statements in the body of the rule are discarded
 *
 * RULE 1:
 *
 * (X, property, Y), (Y, type, B), (B = Bag | B = Seq)
 * (Y, _i, O_i)
 * =>
 * (X, property, O_i), ((X, property, O_i), RDFX.order, i)
 *
 * RULE 2:
 *
 * (Y, property, X), (Y, type, B), (B = Bag | B = Seq)
 * (Y, _i, O_i)
 * =>
 * (O_i, property, X), ((O_i, property, X), RDFX.backwardOrder, i)
 *
 * Notes:
 *
 * translation with Datalog would be difficult, since access to
 * literal values is required (_1, _2 etc.)
 *
 */

public class OrderRDF2Reification extends SimpleModelGateway {

  public void init(Model srcMD, Model destMD) {
  }

  protected void processStatement(Statement st) throws ModelException {

    switch(currentPass()) {
    case 1:
      applyRule1(st);
      break;
    case 2:
      applyRule2(st);
      break;
    }
  }
  
  protected int totalPassNum() {

    return 2;
  }

  boolean applyRule1(Statement st) throws ModelException {

    Resource X = st.subject();
    Resource prop = st.predicate();
    RDFNode Y = st.object();

    // find out whether Y is a Container
    Statement ts = get(Y, RDF.type, RDF.Seq);

    if(ts == null)
      ts = get(Y, RDF.type, RDF.Bag);

    if(ts == null || RDFUtil.getOrd(prop) >= 1)
      return false;

    // Y is a container
    // go over all ordinal properties

    dest.remove(st);
    dest.remove(ts);

    boolean moreThanOne = get(Y, factory().createOrdinal(2), null) != null;

    //    System.err.println("MORE THAN ONE for " + st + " is " + moreThanOne);

    for(int i = 1; ; i++) {

      Statement os = get(Y, factory().createOrdinal(i), null);
      if(os == null)
	break;
      else
	dest.remove(os);
      
      RDFNode Oi = os.object();
	
      Statement rs = stmt(X, prop, Oi);
      dest.add(rs);
      if(moreThanOne) {
	dest.add(stmt(rs, RDFX.order, factory().createLiteral(i)));
      }
    }
    return true;
  }


  boolean applyRule2(Statement st) throws ModelException {

    Resource Y = st.subject();
    Resource prop = st.predicate();
    RDFNode X = st.object();

    // find out whether Y is a Container
    Statement ts = get(Y, RDF.type, RDF.Seq);

    if(ts == null)
      ts = get(Y, RDF.type, RDF.Bag);

    if(ts == null || RDFUtil.getOrd(prop) >= 1)
      return false;

    // Y is a container

    //    System.out.println("--- " + Y);

    // go over all ordinal properties

    dest.remove(st);
    dest.remove(ts);

    boolean moreThanOne = get(Y, factory().createOrdinal(2), null) != null;

    for(int i = 1; ; i++) {

      Statement os = get(Y, factory().createOrdinal(i), null);
      if(os == null)
	break;
      else
	dest.remove(os);
      
      RDFNode Oi = os.object();
	
      if(Oi instanceof Resource) {
	Statement rs = stmt((Resource)Oi, prop, X);
	dest.add(rs);
	if(moreThanOne) {
	  dest.add(stmt(rs, RDFX.backwardOrder, factory().createLiteral(i)));
	}
      }
    }
    return true;
  }
}
