package com.interdataworking.mm.interpreter;

import org.xml.sax.InputSource;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;

import com.interdataworking.*;
import com.interdataworking.mm.*;

import java.util.*;
import java.io.*;
import java.sql.SQLException;

import org.w3c.rdf.model.*;
import org.w3c.rdf.syntax.*;
import org.w3c.rdf.util.*;

import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;

import edu.stanford.db.rdf.query.*;
import edu.stanford.db.rdf.query.memory.*;
import edu.stanford.db.rdf.query.orderings.*;

import org.w3c.rdf.implementation.syntax.sirpac.SiRPAC; // need to set default parsing params

import javax.swing.*;

/**
 * Script or portion of script with optional parameters and local variables
 */

public class ExecutionContext {

  static final int POPUP_DELTA_X = 30;
  static final int POPUP_DELTA_Y = 30;
  int popupX = 50, popupY = 50; // initial coordinates

  static final String WRITESEP = "----------------------------------------";
  static final Object UNDEFINED = new Object();

  public QueryIterator[] Q_D1_ = {
    new FindIterator("Def", Script.operatorName, "N"),
    new FindIterator("Def", RDF.type, Script.Operator)
  };

  public QueryIterator[] Q_D2_ = {
    new FindIterator("Term", Script.operatorName, "N"),
    new NegatedIterator(new JoinIterator(Q_D1_))
  };

  public QueryIterator Q_D = new UniqueIterator(new JoinIterator(Q_D2_), "N");

  public QueryIterator[] Q_G = {
    new FindIterator("Def", Script.operatorName, "N"),
    new FindIterator("Def", RDF.type, Script.Operator),
    new FindIterator("Def", Script.javaClassName, "CLS")
  };

  Collection languageCallbacks = new HashSet();
  Model script;

  File scriptDir; // top script directory
  List directoryStack = new ArrayList(); // of File's, used using parsing
  HashMap processedIncludes = new HashMap(); // include "filename"; -> Script instance (Resource)
  HashMap definitions = new HashMap(); // aliases and operators;  XMI2Obj -> Resource (com.interdataworking....)

  int depth = 0;
  List trace = new ArrayList(); // execution trace; list of resources

  List stack = new ArrayList(); // of variables
  boolean forceReturn = false; // when return statement is encountered, we need to exit operator

  boolean dryRun = false;
  static final String DUMMY = "DUMMY"; // used for dry run

  Map globalConstants = new HashMap();

    IOPlugIn ioPlugIn;
    
    public ExecutionContext() {

	globalConstants.put("null", null);
	globalConstants.put("true", Boolean.TRUE);
	globalConstants.put("false", Boolean.FALSE);
    }

    public void setIOPlugIn(IOPlugIn ioPlugIn) {

	this.ioPlugIn = ioPlugIn;
    }

    void checkIOPlugIn() {
	if(ioPlugIn == null)
	    bailOut("No I/O plug-in registered in ExecutionContext. Use setIOPlugIn() first.");
    }

    public void configDB(String dbDriver, String dbURI, String dbUser, String dbPwd) {

	if(ioPlugIn == null)
	    log("WARNING: no database plug-in registered");
	else
	    ioPlugIn.configDB(dbDriver, dbURI, dbUser, dbPwd);
    }
    
  public ParameterList createList() {

    return new ParameterList(this);
  }

  public ParameterList toParameterList(List l) {

    if(l instanceof ParameterList)
      return (ParameterList)l;

    ParameterList result = createList();
    result.addAll(l);
    return result;
  }

  public void setScriptModel(Model script) {

    this.script = script;
  }

  public Model getScriptModel() {

    return script;
  }

  public void pushd(File f) {

    if(!f.isDirectory())
      f = f.getParentFile(); // enclosing directory
    directoryStack.add(f);
    if(scriptDir == null)
      scriptDir = f;
    //    System.err.println("PUSHING DIR: " + f);
  }

  public File scriptDir() {

    return scriptDir;
  }

  public String replaceSlashes(String str) {

    StringBuffer b = new StringBuffer();
    for(int i=0; i < str.length(); i++) {

      char c = str.charAt(i);
      if(c == '\\' || c == '/')
	c = File.separatorChar;
      b.append(c);
    }
    return b.toString();
  }

  public File getRelative(File f, String str) {

    str = replaceSlashes(str);
    File r = new File(str);
    if(!r.isAbsolute() && f != null)
      r = new File(f, str);
    return r;
  }

  public File currentDir() {

    //    System.err.println("CUR DIR STACK: " + directoryStack);
    File f = null;
    if(directoryStack.size() > 0)
      f = (File)directoryStack.get(directoryStack.size()-1);
    //    System.err.println("CUR DIR: " + f);
    return f;
  }

  public void popd() {

    if(directoryStack.size() > 0)
      directoryStack.remove(directoryStack.size()-1);
    //    System.err.println("POPPED DIR: " + directoryStack);
  }

  public void log(String s) {

    if(!dryRun)
      System.err.print(indent() + s);
  }

  public void error(String s) {

    if(dryRun) {
      System.err.println(indent() + "WARNING: " + s);
      printStackTrace();
    }
    else
      bailOut(s);
  }

  /*
  public void dbRead() {

    log("Reading into " + var.image + " from database " + url + " ... ");
    initDB();
    DBTransaction stmt = db.getDefaultTransaction();
    Model m = db.fetchModel(stmt, factory.createResource(url));

    setVar(var.image, m);
    log("Done " + m.size() + " statements.\n");
  }

  public void dbWrite() {

    if(list.size() != 1)
      throw new ParseException("Refusing to write a multi-valued expression");
    Model m = (Model)list.get(0);

    log("Writing into database " + url + " ... ");

    initDB();
    DBTransaction stmt = db.getDefaultTransaction();
    db.storeModel(stmt, url, m);

    log("Done.\n");
  }
  */

  public void addProcessed(String fileName, Resource script) {

    processedIncludes.put(fileName, script);
  }

  public boolean isProcessed(String fileName) {

    return processedIncludes.containsKey(fileName);
  }

  public Resource getProcessed(String fileName) {

    return (Resource)processedIncludes.get(fileName);
  }

  public void addDefinition(String operator, Resource r) {

    definitions.put(operator, r);
  }

  public boolean isDefined(String operator) {

    return definitions.containsKey(operator);
  }

  public Resource getDefinition(String operator) {

    return (Resource)definitions.get(operator);
  }

  // convenience method
  public Resource createUniqueResource() throws ModelException {

    return script.getNodeFactory().createUniqueResource();
  }

  public Literal createLiteral(String s) throws ModelException {

    return script.getNodeFactory().createLiteral(s);
  }

  public void add(Resource s, Resource p, RDFNode o, int pos) throws ModelException {

    Statement st = script.getNodeFactory().createStatement(s, p, o);
    script.add(st);
    if(pos > 0)
      add(st, RDFX.order, String.valueOf(pos));
  }

  public void add(Resource s, Resource p, RDFNode o) throws ModelException {

    add(s, p, o, -1);
  }

  public void add(Resource s, Resource p, String o, int pos)  throws ModelException {

    add(s, p, createLiteral(o), pos);
  }

  public void add(Resource s, Resource p, String o)  throws ModelException {

    add(s, p, o, -1);
  }

  public void addOrdered(Resource s, Resource p, List list) throws ModelException {

    for(int i=0; i < list.size(); i++) {

      Object obj = list.get(i);
      if(obj instanceof String)
        add(s, p, (String)obj, i+1);
      else
        add(s, p, (RDFNode)obj, i+1);
    }
  }

  public void validate(Resource start) throws ModelException {

    /*
    for(NodeIterator it = new NodeResultSet(Q_D, script); it.next(); ) {

      log("WARNING: undefined operator " + it.get("N") + "\n");
    }

    for(NodeIterator it = new NodeResultSet(Q_G, script); it.next(); ) {

      String cls = String.valueOf(it.get("CLS"));
      try {
	Object fct = createInstance(cls);
	if(!(fct instanceof UntypedGateway))
	  log("WARNING: operator " + it.get("N") + " with alias " + it.get("CLS") + " does not implement UntypedGateway interface\n");
      } catch (Exception any) {
	log("WARNING: cannot load class " + it.get("CLS") + " for operator " + it.get("N") + "\n");
      }
    }
    */

    dryRun = true;
    run(start);
    dryRun = false;
  }

  // returns a list only for return statement
  public List run(Resource r) throws ModelException {

    try {
      enterBlock();
      List result = execute(r);
      exitBlock();
      return result;

    } catch(Exception any) {

      any.printStackTrace(System.err);
      bailOut("Script execution error");
      return null;
    }
  }


  protected ParameterList executeList(List l) throws ModelException  {

    for(Iterator it = l.iterator(); it.hasNext();) {

      ParameterList res = execute((Resource)it.next());
      if(forceReturn) // return statement reached
	return res;
    }
    return createList();
  }

  // returns a list only for return statement
  protected ParameterList execute(Resource r) throws ModelException {

    pushTrace(r);

    ParameterList out = createList();

    if(RDFUtil.isInstanceOf(script, r, Script.Script)) {
      // evaluate each expression
      evaluateEach(getOrdered(r, Script.expression));

    } else if(RDFUtil.isInstanceOf(script, r, Script.Assignment)) {

      List vars = getOrdered(r, Script.variable);
      Resource term = RDFUtil.getObjectResource(script, r, Script.term);
      List values = evaluateTerm(term);
      assignVars(vars, values);

    } else if(RDFUtil.isInstanceOf(script, r, Script.ReturnExpression)) {

      List terms = getOrdered(r, Script.term);
      if(terms != null)
	out = evaluateTerms(terms);
      else
        out = createList(); // empty return
      forceReturn = true;

    } else if(RDFUtil.isInstanceOf(script, r, Script.IfStatement)) {
      
      Resource ifExpression = RDFUtil.getObjectResource(script, r, Script.ifExpression);
      ParameterList values = evaluateTerm(ifExpression);
      
      Resource nextTerm = null;
      if(getBoolean(values))
	out = executeList(getOrdered(r, Script.thenStatement));
      else
	out = executeList(getOrdered(r, Script.elseStatement));

// 	nextTerm = RDFUtil.getObjectResource(script, r, Script.thenStatement);
//       else
// 	nextTerm = RDFUtil.getObjectResource(script, r, Script.elseStatement);
      
//       if(nextTerm != null)
// 	out = execute(nextTerm);

    } else if(RDFUtil.isInstanceOf(script, r, Script.DoStatement)) {
      
      Resource ifExpression = RDFUtil.getObjectResource(script, r, Script.ifExpression);

      ParameterList values = null;

      do {
	
	evaluateEach(getOrdered(r, Script.expression));
	values = evaluateTerm(ifExpression);
      
      } while(!dryRun && getBoolean(values));


    } else if(RDFUtil.isInstanceOf(script, r, Script.WhileStatement)) {
      
      Resource ifExpression = RDFUtil.getObjectResource(script, r, Script.ifExpression);

      while(true) {

	ParameterList values = evaluateTerm(ifExpression);
	if(dryRun || !getBoolean(values))
	  break;
	
	evaluateEach(getOrdered(r, Script.expression));
      }


    } else if(RDFUtil.isInstanceOf(script, r, Script.WriteExpression)) {

      writeStmt(r);

    } else if(RDFUtil.isInstanceOf(script, r, Script.ReadExpression)) {

      readStmt(r);
    }

    popTrace();
    return out;
  }

  protected Model createModel() throws ModelException {

    return new RDFFactoryImpl().createModel();
  }


  protected void readStmt(Resource r) throws ModelException {

    String varName = RDFUtil.getObjectLiteral(script, r, Script.variable);
    String streamName = RDFUtil.getObjectLiteral(script, r, Script.streamName);
    Resource streamTarget = RDFUtil.getObjectResource(script, r, Script.streamTarget);

    String streamTargetStr = null;
    if(streamTarget != null) {
      ParameterList out = evaluateTerm(streamTarget);
      if(out != null && out.size() > 0)
	streamTargetStr = (String)out.get(0);
    }

    Model m = null;

    try {

      if("database".equals(streamName)) {

	// read from database
	log("Reading from database under name " + streamTargetStr + " ...\n");
      
	if(!dryRun) {

	    checkIOPlugIn();
            m = ioPlugIn.readModel(streamTargetStr);
      	}
      
	//	log("Done.\n");
	
      } else { // file or stream name

	// determine parser

	String cls = RDFUtil.getObjectLiteral(script, r, Script.aliasName);
	Object fct = createInstance(cls);
	if(!(fct instanceof RDFParser))
	  error("" + cls + " instance " + fct + " does not implement RDFParser");
	RDFParser p = (RDFParser)fct;

	if(p instanceof SiRPAC)
	  ((SiRPAC)p).setGenerateUniqueResources(true);

	// determine target: stream, file, or database

	if(!dryRun) {

	  Reader in = null;

	  if("System.in".equals(streamName))
	    in = new InputStreamReader(System.in);

	  else if("file".equals(streamName)) {

	    if(streamTargetStr != null) {
	      log("Reading from file " + streamTargetStr + " ...\n");
	      in = new FileReader(getRelative(scriptDir(), streamTargetStr));
	    }
	  }

	  if(in == null)
	    error("Internal error: invalid stream name for reading: " + streamName + ", " + streamTargetStr);
	  
	  InputSource source = new InputSource(in);
	  m = createModel();
	  p.parse(source, new ModelConsumer(m));
	  if("file".equals(streamName))
	    in.close();
	}
      }
    } catch (Exception any) {

//       if(!(any instanceof IOException))
// 	any.printStackTrace(System.err);
      log("WARNING: Read failed, variable " + varName + " set to empty model. " + any + "\n");
      m = createModel();
      //      error("Could not read: " + any);
    }

    setVar(varName, m);
  }

  protected void writeStmt(Resource r) throws ModelException {

    Resource term = RDFUtil.getObjectResource(script, r, Script.term);
    ParameterList values = evaluateTerm(term);

    //    System.err.println("--- WRITE: " + r + ", values=" + values);

    if(values.size() != 1)
      log("Writing only the first parameter of a multi-valued expression");

    Model m = dryRun ? null : (Model)values.get(0);

    String streamName = RDFUtil.getObjectLiteral(script, r, Script.streamName);
    Resource streamTarget = RDFUtil.getObjectResource(script, r, Script.streamTarget);

    String streamTargetStr = null;
    if(streamTarget != null) {
      ParameterList out = evaluateTerm(streamTarget);
      if(out != null && out.size() > 0)
	streamTargetStr = (String)out.get(0);
    }

    try {

      if("database".equals(streamName)) {

	// write to database
	log("Writing into database under name " + streamTargetStr + " ...\n");

	if(!dryRun) {
	    checkIOPlugIn();
	    ioPlugIn.writeModel(streamTargetStr, m);
	}

	//	log("Done.\n");

      } else {

	// determine serializer

	String cls = RDFUtil.getObjectLiteral(script, r, Script.aliasName);
	Object fct = createInstance(cls);
	if(!(fct instanceof RDFSerializer))
	  bailOut("" + cls + " instance " + fct + " does not implement RDFSerializer");
	RDFSerializer p = (RDFSerializer)fct;

	// determine target: stream, file, or database

	if(!dryRun) {

	  Writer out = null;
	  boolean regularStream = false;
	  
	  if(streamName != null) {
	    //	    log("Writing model of size " + m.size() + " to stream\n");
	  }

	  if("System.out".equals(streamName)) {

	    out = new OutputStreamWriter(System.out);
	    regularStream = true;

	  } else if("System.err".equals(streamName)) {

	    out = new OutputStreamWriter(System.err);
	    regularStream = true;

	  } else if("window".equals(streamName)) {

	    out = new StringWriter();

	  } else if("file".equals(streamName)) {

	    if(streamTargetStr != null) {
	      log("Writing to file " + streamTargetStr + " ...\n");
	      out = new FileWriter(getRelative(scriptDir(), streamTargetStr));
	    }
	  }

	  if(out == null)
	    error("Internal error: invalid stream name for writing: " + streamName + ", " + streamTargetStr);
	  
	  if(out != null) {
	    if(regularStream)
	      log(WRITESEP + "\n");
	    p.serialize(m, out);
	    out.flush();
	    if(regularStream)
	      log(WRITESEP + "\n");
	    else
	      out.close();
	  }

	  if("window".equals(streamName))
	    popupTextWindow(out.toString(), streamTargetStr);
	}
      }

    } catch (Exception any) {

      if(!(any instanceof IOException))
	any.printStackTrace(System.err);
      error("Could not write: " + any);
    }
  }

  // flattens out result
  protected ParameterList evaluateTerms(List terms) throws ModelException  {

    ParameterList result = createList();

    for(Iterator it = terms.iterator(); it.hasNext();) {

      Resource t = (Resource)it.next();
      List values = evaluateTerm(t);
      result.addAll(values);
    }

    return result;
  }

  protected boolean getBoolean(ParameterList l) {

    if(l == null || l.isEmpty())
      return false;
    Object v = l.get(0);
    return (v instanceof Boolean) && ((Boolean)v).booleanValue();
  }

  public ParameterList executeOperator(String opName, List inputList) throws ModelException {

    ParameterList result = createList();

      Resource opDef = (Resource)definitions.get(opName);
      if(opDef == null)
	error("Undefined operator: " + opName);

      // java-defined or explicitly defined?

      String className = RDFUtil.getObjectLiteral(script, opDef, Script.javaClassName);
      if(className != null) {

	// java-defined operator

	log(opName + "()\n");

	Object obj = createInstance(className);
	if(obj instanceof UntypedGateway) {

	  if(!dryRun) {

	    UntypedGateway g = (UntypedGateway)obj;
	    if(inputList.size() < g.getMinInputLen() ||
	       inputList.size() > g.getMaxInputLen())
	      error("Invalid number of input parameters for operator " + opDef + "=" + g);

	    result = toParameterList(g.execute(inputList));

	  } else
	    result = createList();

	} else
	  error("Operator " + opDef + "=" + obj + " does not implement UntypedGateway interface");

      } else {

	// explicitly defined operator

	log(opName + "() :=\n");

	List vars = getOrdered(opDef, Script.variable);
	enterBlock();
	assignVars(vars, inputList); // put variables on stack
	List expr = getOrdered(opDef, Script.expression);
	result = executeList(expr);
	exitBlock(); // exit from subprocedure
	forceReturn = false;
      }

      return result;
  }

  protected ParameterList evaluateTerm(Resource t) throws ModelException  {

    pushTrace(t);
    ParameterList result = createList();

    if(RDFUtil.isInstanceOf(script, t, Script.FunctionalTerm)) {

      List terms = getOrdered(t, Script.term);
      List inputList = evaluateTerms(terms);

      String opName = RDFUtil.getObjectLiteral(script, t, Script.operatorName);
      result = executeOperator(opName, inputList);

    } else if(RDFUtil.isInstanceOf(script, t, Script.Variable)) {

      String varName = RDFUtil.getObjectLiteral(script, t, Script.variableName);
      Object obj = getVar(varName);
      if(obj == UNDEFINED) {
	obj = null;
	log("WARNING: undefined variable " + varName + "\n");
      }
      result = createList();
      result.add(obj);


    } else if(RDFUtil.isInstanceOf(script, t, Script.Constant)) {

      String constValue = RDFUtil.getObjectLiteral(script, t, Script.constantValue);
      if(constValue == null)
	error("Internal error: missing constant name");
      result = createList();
      result.add(constValue);
    }    

    popTrace();
    return result;
  }

  protected void enterBlock() {

    stack.add(new HashMap());
  }

  protected void exitBlock() {

    if(stack.size() > 0)
      stack.remove(stack.size()-1);
  }

  protected void assignVars(List out, List in) {

    if(in == null || out == null) // in=null can happen if nothing is returned from a function
      return;

    if(dryRun) {
      for(int i=0; i < out.size(); i++)
	setVar(String.valueOf(out.get(i)), null);
    } else {
      //      System.err.println("=== ASSIGNMENT: " + out + " := " + in);
      for(int i=0; i < Math.min(in.size(), out.size()); i++)
	setVar(String.valueOf(out.get(i)), in.get(i));
    }
  }

  protected void setVar(String varName, Object value) {

    if(globalConstants.containsKey(varName))
      error("Invalid variable assignment to " + varName);

    Map vars = getVariables();
    if(value != UNDEFINED)
      vars.put(varName, value);
  }

  protected Object getVar(String varName) {

    if(globalConstants.containsKey(varName))
      return globalConstants.get(varName);

    Map vars = getVariables();

    if(!vars.containsKey(varName))
      return UNDEFINED;

    return vars.get(varName);
  }

  protected Map getVariables() {

    return (Map)stack.get(stack.size() - 1);
  }

  protected void evaluateEach(Collection c) throws ModelException  {

    for(Iterator it = c.iterator(); it.hasNext();) {

      Resource r = (Resource)it.next();
      execute(r);
    }
  }

  protected List getOrdered(Resource s, Resource p) throws ModelException  {

    QueryIterator Q = new OrderIterator(new FindIterator("S", "P", "O", "ST"),
					new ByOrderProperty("ST"));

    NodeIterator it = new NodeResultSet(Q, script);
    it.set("S", s);
    it.set("P", p);

    List result = new ArrayList();

    while(it.next())
      result.add(it.get("O"));

    return result;
  }

  void popupTextWindow(String content, String title) {

    JFrame frame = new JFrame();
    JDialog dialog = new JDialog(frame, title, true);
    JTextArea text = new JTextArea(content);
    JScrollPane scroll = new JScrollPane(text);
    dialog.getContentPane().add(scroll);
    //    JPanel panel = new JPanel();
    //    dialog.setSize(800, 600); // dialog.getHeight());
    dialog.setLocation(popupX, popupY);
//     popupX += POPUP_DELTA_X;
//     popupY += POPUP_DELTA_Y;
//    dialog.setVisible(true);
    dialog.pack();
    dialog.show();
//     int w = Math.min(dialog.getWidth(), 600);
//     int h = Math.min(dialog.getHeight(), 400);
    int w = Math.max(dialog.getWidth(), 50);
    int h = Math.max(dialog.getHeight(), 50);
    dialog.setSize(w, h);
//     dialog.invalidate();
//     dialog.repaint();
  }

  String indent() {

    return indent(depth);
  }

  public void pushTrace(Resource t) {

    trace.add(t);
    incDepth();
  }

  public void popTrace() {

    decDepth();
    trace.remove(trace.size()-1);
  }

  public void incDepth() {

    depth++;
  }

  public void decDepth() {

    depth--;
  }

  String indent(int depth) {

    StringBuffer b = new StringBuffer();
    String dbgInfo = getCurrentDebugInfo();
    b.append(dbgInfo);
    if(!"".equals(dbgInfo))
      b.append(':');
    for(int i = 0; i < depth; i++)
      b.append("   ");
    return b.toString();
  }

  String getCurrentDebugInfo() {

    return trace.isEmpty() ? "" : getDebugInfo((Resource)trace.get(trace.size()-1));
  }

  String getDebugInfo(Resource t) {

    String r = null;
    try {
      r = RDFUtil.getObjectLiteral(script, t, Script.debugInfo);
    } catch (ModelException any) {}

    return r == null ? "" : r;
  }

  void printStackTrace() {

    for(int i = trace.size(); --i >=0;) {

      Resource t = (Resource)trace.get(i);
      System.err.println("\tat " + t + " " + getDebugInfo(t));
    }
  }

  void bailOut(String msg) {

      if(ioPlugIn != null)
	  ioPlugIn.cleanup();

    printStackTrace();
    throw new RuntimeException(msg);
    //    System.err.println(msg);
    //    System.exit(1);
  }


  public Collection getLanguageCallbacks() {

    return new ArrayList(languageCallbacks);
  }

  public LanguageCallback getLanguageCallback(Model m) throws ModelException {

    for(Iterator it = languageCallbacks.iterator(); it.hasNext();) {

      LanguageCallback cb = (LanguageCallback)it.next();
      if(cb.isCallbackFor(m))
        return cb;
    }
    return null;
  }

  Object createInstance(String cname) throws ModelException {

    Resource alias = (Resource)definitions.get(cname);
    if(alias != null) {

      cname = RDFUtil.getObjectLiteral(script, alias, Script.javaClassName);
    }

    Class c = null;
    try {
       c = Class.forName(cname);
    } catch (Exception any) {
      bailOut("Undefined alias or class name: " + cname);
    }

    try {
      return c.newInstance();
    } catch (Exception ex) {
      bailOut("Cannot create instance of " + c + "( " + cname + "): " + ex);
      return null; // not reached
    }
  }

  public void addLanguageCallback(String callback) throws ModelException {

    Object cb = createInstance(callback);
    if(!(cb instanceof LanguageCallback))
      log("Invalid language callback " + cb + ", not an instance of LanguageCallback!\n");
    else
      log("Registering: " + cb + "\n");

    languageCallbacks.add(cb);
  }

}
