/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package edu.stanford.db.rdf.query.memory;

import org.w3c.rdf.implementation.model.ResourceImpl;
import org.w3c.rdf.model.*;
import edu.stanford.db.rdf.query.*;
import org.w3c.rdf.util.RDFUtil;
import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;
import edu.stanford.db.rdf.query.util.Skolem;

import java.util.*;

/**
 * This is a query iterator that sets a flag whether a given
 * variable is seen for the first time
 * If flagVarName is not set, repetitions are tacitly skipped,
 * otherwise variable flagVarName will be set to UNIQUE or SKIP
 *
 * @author Sergey Melnik <melnik@db.stanford.edu> */

public class GroupByIterator implements QueryIterator {

  public static final Resource UNIQUE = new ResourceImpl("GroupByIterator:unique");
  public static final Resource SKIP = new ResourceImpl("GroupByIterator:skip");
  static final Resource NULL = new ResourceImpl("GroupByIterator:NULL");

  QueryIterator it;
  String[] varNames; // String uniqueVarName;
  String flagVarName;
  HashSet doneValues = new HashSet();

  public GroupByIterator(String var, QueryIterator it) {

    this(ConditionIterator.oneStringArray(var), null, it);
  }

  public GroupByIterator(String[] varNames, QueryIterator it) {

    this(varNames, null, it);
  }

  public GroupByIterator(String var, String flagVarName, QueryIterator it) {

    this(ConditionIterator.oneStringArray(var), flagVarName, it);
  }

  public GroupByIterator(String[] varNames, String flagVarName, QueryIterator it) {

    this.it = it;
    this.varNames = varNames;
    this.flagVarName = flagVarName;
    Object varValues = new Object[varNames.length];
  }

  public void reset(Hashtable vars, Model model) throws ModelException {

    it.reset(vars, model);
    doneValues = new HashSet();
  }

  // returns true if match found (vars updated if necessary)
  public boolean next(Hashtable vars) throws ModelException {

    boolean tryAgain = true;

    do {
      if(!it.next(vars))
	return false;

      Object[] varValues = new Object[varNames.length];
      boolean allNull = true;
      for(int i=0; i < varNames.length; i++) {
	Object o = vars.get(varNames[i]);
	varValues[i] = o != null ? o : NULL;
	if(o != null)
	  allNull = false;
      }

      if(allNull)
	return true;

      Skolem skolem = new Skolem(varValues);

      boolean inDoneValues = doneValues.contains(skolem);
      tryAgain = inDoneValues && flagVarName == null;
      doneValues.add(skolem);
      if(flagVarName != null)
	vars.put(flagVarName, inDoneValues ? SKIP : UNIQUE);
    } while(tryAgain);

    return true;
  }

  // returns names of the variables used
  public Enumeration getVariables() throws ModelException {

    return it.getVariables();
  }

  // returns execution cost of the iterator (e.g. stylizised number
  // of instructions/method calls)
  public int getCost() {

    return it.getCost();
  }

  public QueryIterator duplicate() {

    GroupByIterator oi = new GroupByIterator(varNames, flagVarName, it.duplicate());
    oi.doneValues = (HashSet)doneValues.clone();
    return oi;
  }

}
