/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package edu.stanford.db.rdf.query.memory;

import org.w3c.rdf.model.*;
import edu.stanford.db.rdf.query.*;
import java.util.*;

/**
 * This is a query iterator that orders the result set.
 * The variable by which the order is established must deliver statements.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu> */

public class OrderIterator implements QueryIterator {
    
  QueryIterator it;
  Object[] mapList; // array of maps (Hashtables)
  Model m;
  int current;
  int cost;
  Hashtable initVars;

  ModelException savedException;
  SubstitutionComparator sc;
  Comparator comparator;

  public OrderIterator(QueryIterator it, SubstitutionComparator scomp) {

    this.it = it;
    this.sc = scomp;
    this.comparator = new Comparator() {

	public int compare(Object a, Object b) {

	  cost++;
	  try {
	    //	    System.err.println("==-=-=-=- model is " + OrderIterator.this.m + " IN " + OrderIterator.this);
	    return sc.compare((Map)a, (Map)b, OrderIterator.this.m);

	  } catch (ModelException any) {

	    savedException = any;
	    return 0;
	  }
	}

	public boolean equals(Object that) {
	  return this == that;
	}
      };
  }

  /*
  protected OrderIterator(QueryIterator it, Comparator comparator) {

    this.it = it;
    this.comparator = comparator;
  }
  */

  public void reset(Hashtable vars, Model model) throws ModelException {

    //    System.err.println("======== RESETTING ORDER ITERATOR " + this + " with " + model);

    it.reset(this.initVars = vars, model);
    this.m = model;
    mapList = null;
    cost = 0;
  }

  // returns true if match found (vars updated if necessary)
  public boolean next(Hashtable vars) throws ModelException {

    init();

    if(current < mapList.length) {

      Hashtable t = (Hashtable)mapList[current++];
      // copy the content of the hashtable - inefficient. TODO: change interface?
      vars.clear();
      vars.putAll(t);
      /*
      for(Enumeration en = t.keys(); en.hasMoreElements();) {

	Object varName = en.nextElement();
	Object value = t.get(varName);
	vars.put(varName, value);
      }
      */
      return true;
    } else
      return false;
  }

  // returns names of the variables used
  public Enumeration getVariables() throws ModelException {

    return it.getVariables();
  }

  // returns execution cost of the iterator (e.g. stylizised number
  // of instructions/method calls)
  public int getCost() {

    return it.getCost() + cost;
  }

  public QueryIterator duplicate() {

    OrderIterator oi = new OrderIterator(it.duplicate(), sc);
    //    oi.sc = sc;
    oi.mapList = mapList;
    oi.m = m;
    oi.current = current;
    //    System.err.println("====== DUPLICATING ORDER ITERATOR " + this + " INTO " + oi + " WITH model=" + oi.m);
    return oi;
  }

  /* computes mapList and sorts it */

  void init() throws ModelException {
    
    if(mapList == null) {

      ArrayList list = new ArrayList();

      boolean hasMore;
      Hashtable h = (Hashtable)initVars.clone();
      while(true) {
	hasMore = it.next(h);
	if(hasMore) {
	  list.add(h.clone());
	} else
	  break;
      }

      cost = 0;
      current = 0;
      mapList = list.toArray();
      savedException = null;

      //      System.err.println("====== SORTING: " + list + " USING " + sc + ", MODEL=" + m + " IN " + this);

      Arrays.sort(mapList, comparator);

      if(savedException != null) {
	savedException.printStackTrace(System.err);
	throw savedException;
      }
    }
  }

}
